"""
This module contains code that can be used to run automated tests of
datapackage pipelines processors

It is used internally to test the standard library processors but supports
external use as well.
"""
import os
import subprocess
import sys
import json


class ProcessorFixtureTestsBase(object):

    def __init__(self, fixtures_path):
        self._fixtures_path = fixtures_path

    def get_tests(self):
        for dirpath, _, filenames in os.walk(self._fixtures_path):
            for filename in filenames:
                data_in, data_out, dp_out, params, processor = \
                    self._load_fixture(dirpath, filename)

                def inner(processor_, params_, data_in_, dp_out_, data_out_):
                    def inner2():
                        return self._test_single_fixture(
                            processor_, params_,
                            data_in_, dp_out_,
                            data_out_,
                            self._get_procesor_env())
                    return inner2
                yield filename, inner(processor, params,
                                      data_in, dp_out, data_out)

    def _get_procesor_env(self):
        """
        Set the environment variables for the sub-process that runs the
        processor extending classes will most likely want to set the PYTHONPATH
        variable to correct value
        """
        return {}

    def _get_processor_file(self, processor):
        """
        Must be implemented by extending classes to return the path to the
        processor python file.

        :param processor: the value of the 1st part of the fixture (the name of
            the processor to test)
        :return: full path to the processor python file
        """
        raise NotImplementedError()

    def _load_fixture(self, dirpath, filename):
        with open(os.path.join(dirpath, filename), encoding='utf8') as f:
            parts = f.read().split('--\n')
        processor, params, dp_in, data_in, dp_out, data_out = parts
        processor_file = self._get_processor_file(processor)
        params = rejsonize(params)
        dp_out = rejsonize(dp_out)
        dp_in = rejsonize(dp_in)
        data_in = (dp_in + '\n\n' + data_in).encode('utf8')
        return data_in, data_out, dp_out, params, processor_file

    def _run_processor(self, processor, parameters, data_in, env):
        '''Run the passed `processor` and return the output'''
        process = subprocess.run([sys.executable, processor, '1',
                                  parameters, 'False', ''],
                                 input=data_in,
                                 stdout=subprocess.PIPE,
                                 env=env)
        return process.stdout.decode('utf8')

    def _test_single_fixture(self, processor, parameters, data_in,
                             dp_out, data_out, env):
        """Test a single processor with the given fixture parameters"""
        output = self._run_processor(processor, parameters, data_in, env)
        self.test_fixture(output, dp_out, data_out)

    @staticmethod
    def test_fixture(processor_output, dp_out, data_out):
        '''Receives processor output and performs standard tests. Can be
        overridden in subclasses.'''
        (actual_dp, *actual_data) = processor_output.split('\n\n', 1)
        assert actual_dp == dp_out, \
            "unexpected value for output datapackage: {}".format(actual_dp)
        if len(actual_data) > 0:
            actual_data = actual_data[0]
            actual_data = actual_data.split('\n')
            expected_data = data_out.split('\n')
            assert len(actual_data) == len(expected_data), \
                "unexpected number of output lines: {}, actual_data = {}" \
                .format(len(actual_data), actual_data)
            for line_num, (actual, expected) in enumerate(zip(actual_data,
                                                              expected_data)):
                line_msg = "output line {}".format(line_num)
                if len(actual) == 0:
                    assert len(expected) == 0, \
                        "{}: did not get any data (but expected some)" \
                        .format(line_msg)
                else:
                    rj_actual = rejsonize(actual)
                    assert rj_actual == rejsonize(expected), \
                        "{}: unexpected data: {}".format(line_msg, rj_actual)


def rejsonize(s):
    return json.dumps(json.loads(s), sort_keys=True, ensure_ascii=True)



import os

import yaml

from ..errors import SpecError
from .base_parser import BaseParser, PipelineSpec


class BasicPipelineParser(BaseParser):

    SPEC_FILENAME = 'pipeline-spec.yaml'

    @classmethod
    def check_filename(cls, filename):
        return filename == cls.SPEC_FILENAME

    @classmethod
    def to_pipeline(cls, fullpath):
        dirpath = os.path.dirname(fullpath)
        with open(fullpath, encoding='utf8') as spec_file:
            try:
                spec = yaml.load(spec_file.read())
                for pipeline_id, pipeline_details in spec.items():
                    pipeline_id = os.path.join(dirpath, pipeline_id)
                    yield PipelineSpec(path=dirpath,
                                       pipeline_id=pipeline_id,
                                       pipeline_details=pipeline_details)
            except yaml.YAMLError as e:
                error = SpecError('Invalid Pipeline Spec', str(e))
                yield PipelineSpec(path=dirpath, errors=[error])



import os

import yaml

from ..resolver import resolve_generator
from ..errors import SpecError
from .base_parser import BaseParser, PipelineSpec


class SourceSpecPipelineParser(BaseParser):

    SOURCE_FILENAME_SUFFIX = '.source-spec.yaml'

    @classmethod
    def check_filename(cls, filename):
        return filename.endswith(cls.SOURCE_FILENAME_SUFFIX)

    @classmethod
    def to_pipeline(cls, fullpath):
        filename = os.path.basename(fullpath)
        dirpath = os.path.dirname(fullpath)

        module_name = filename[:-len(cls.SOURCE_FILENAME_SUFFIX)]
        generator = resolve_generator(module_name)

        if generator is None:
            message = 'Unknown source description kind "{}" in {}' \
                .format(module_name, fullpath)
            error = SpecError('Unknown source kind', message)
            yield PipelineSpec(path=dirpath, errors=[error])
            return

        try:
            with open(fullpath, encoding='utf8') as spec_file:
                source_spec = yaml.load(spec_file.read())
                if generator.internal_validate(source_spec):
                    spec = generator.internal_generate(source_spec)
                    for pipeline_id, pipeline_details in spec:
                        pipeline_id = os.path.join(dirpath, pipeline_id)
                        yield PipelineSpec(path=dirpath,
                                           pipeline_id=pipeline_id,
                                           pipeline_details=pipeline_details,
                                           source_details=source_spec)
                else:
                    message = 'Invalid source description for "{}" in {}' \
                        .format(module_name, fullpath)
                    error = SpecError('Invalid Source', message)
                    yield PipelineSpec(path=dirpath, errors=[error])

        except yaml.YAMLError as e:
            error = SpecError('Invalid Source Spec', str(e))
            yield PipelineSpec(dirpath, None, None, None, [error])



import logging

from datapackage_pipelines.status import status
from .celery_app import celery_app
from ..specs import pipelines, PipelineSpec
from ..manager.tasks import execute_pipeline


def trigger_dirties(completed=None, trigger=None):
    for spec in pipelines():
        pipeline_id = spec.pipeline_id
        if (len(spec.errors) == 0 and
                (completed is None or completed in spec.dependencies) and
                (spec.dirty or
                 status.is_waiting(pipeline_id) or
                 trigger == 'schedule')):
            status.register(spec.pipeline_id,
                            spec.cache_hash,
                            spec.pipeline_details,
                            spec.source_details,
                            spec.errors)
            logging.info('Executing DIRTY task %s', pipeline_id)
            pipeline_status = status.queued(pipeline_id)
            execute_pipeline_task.delay(pipeline_id,
                                        spec.pipeline_details,
                                        spec.path,
                                        'dirty-task' if trigger is None else trigger,
                                        pipeline_status.data['queued'])


@celery_app.task
def execute_scheduled_pipeline(pipeline_id):
    for spec in pipelines():
        if spec.pipeline_id == pipeline_id:
            if len(spec.errors) == 0:
                status.register(spec.pipeline_id,
                                spec.cache_hash,
                                spec.pipeline_details,
                                spec.source_details,
                                spec.errors)
                logging.info('Executing SCHEDULED task %s', pipeline_id)
                execute_pipeline_task.delay(pipeline_id,
                                            spec.pipeline_details,
                                            spec.path,
                                            'scheduled',
                                            0)
            else:
                logging.warning('Skipping SCHEDULED task %s, as it has errors %r', pipeline_id, spec.errors)
            break


@celery_app.task
def execute_pipeline_task(pipeline_id,
                          pipeline_details,
                          pipeline_cwd,
                          trigger,
                          queue_time):

    last_queue_time = status.get_status(pipeline_id).get('queued')
    if queue_time in {0, last_queue_time}:
        spec = PipelineSpec(pipeline_id=pipeline_id,
                            pipeline_details=pipeline_details,
                            path=pipeline_cwd)
        success, _ = \
            execute_pipeline(spec,
                             trigger,
                             False)

        if success:
            trigger_dirties(pipeline_id, trigger)



import os
import logging

import itertools

import tabulator

from jsontableschema import Schema

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher
from datapackage_pipelines.utilities.tabulator_txt_parser import TXTParser


def _reader(opener, _url):
    yield None
    filename = os.path.basename(_url)
    logging.info('%s: OPENING %s', filename, _url)
    _schema, _headers, _reader, _close = opener()
    num_headers = len(_headers)
    i = 0
    for i, row in enumerate(_reader):
        row = [str(x).strip() for x in row]
        values = set(row)
        if len(values) == 1 and '' in values:
            # In case of empty rows, just skip them
            continue
        output = dict(zip(_headers, _schema.cast_row(row[:num_headers])))
        yield output

        i += 1
        if i % 10000 == 0:
            logging.info('%s: %d rows', filename, i)
            # break
    _close()
    logging.info('%s: TOTAL %d rows', filename, i)


def dedupe(headers):
    _dedupped_headers = []
    headers = list(map(str, headers))
    for hdr in headers:
        if hdr is None:
            continue
        hdr = hdr.strip()
        if len(hdr) == 0:
            continue
        if hdr in _dedupped_headers:
            i = 0
            deduped_hdr = hdr
            while deduped_hdr in _dedupped_headers:
                i += 1
                deduped_hdr = '%s_%s' % (hdr, i)
            hdr = deduped_hdr
        _dedupped_headers.append(hdr)
    return _dedupped_headers


def row_skipper(rows_to_skip):
    def _func(extended_rows):
        for number, headers, row in extended_rows:
            if number > rows_to_skip:
                yield (number-rows_to_skip, headers, row)
    return _func


def stream_reader(_resource, _url, _ignore_missing):
    def get_opener(__url, __resource):
        def opener():
            _params = dict(headers=1)
            _params.update(
                dict(x for x in __resource.items()
                     if x[0] not in {'path', 'name', 'schema',
                                     'mediatype', 'skip_rows'}))
            skip_rows = __resource.get('skip_rows', 0)
            if _params.get("format") == "txt":
                # datapackage-pipelines processing requires having a header row
                # for txt format we add a single "data" column
                _params["headers"] = ["data"]
                _params["custom_parsers"] = {"txt": TXTParser}
                _params["allow_html"] = True
            _stream = tabulator.Stream(__url, **_params,
                                       post_parse=[row_skipper(skip_rows)])
            try:
                _stream.open()
                _headers = dedupe(_stream.headers)
                _schema = __resource.get('schema')
                if _schema is not None:
                    _schema = Schema(_schema)
                return _schema, _headers, _stream, _stream.close
            except tabulator.exceptions.TabulatorException as e:
                logging.warning("Error while opening resource from url %s: %r",
                                _url, e)
                if not _ignore_missing:
                    raise
                return {}, [], [], lambda: None
        return opener

    schema, headers, stream, close = get_opener(url, _resource)()
    if schema is None:
        schema = {
            'fields': [
                {'name': header, 'type': 'string'}
                for header in headers
                ]
        }
        _resource['schema'] = schema

    close()
    del stream

    return itertools\
        .islice(
            _reader(
                get_opener(_url, _resource),
                _url),
            1, None)


parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
ignore_missing = parameters.get('ignore-missing', False)

new_resource_iterator = []
for resource in datapackage['resources']:

    path = resource.get('path')
    if path is not None and '://' not in path:
        new_resource_iterator.append(next(resource_iterator))
    elif 'url' in resource:
        url = resource['url']

        name = resource['name']
        if not resources.match(name):
            continue

        path = os.path.join('data', name + '.csv')
        resource['path'] = path

        del resource['url']
        rows = stream_reader(resource, url, ignore_missing or url == "")
        new_resource_iterator.append(rows)

spew(datapackage, new_resource_iterator)



import os
from datapackage_pipelines.utilities.lib_test_helpers import ProcessorFixtureTestsBase

ROOT_PATH = os.path.join(os.path.dirname(__file__), '..', '..')
ENV = os.environ.copy()
ENV['PYTHONPATH'] = ROOT_PATH


class StdlibfixtureTests(ProcessorFixtureTestsBase):

    def _get_procesor_env(self):
        return ENV

    def _get_processor_file(self, processor):
        processor = processor.replace('.', '/')
        return os.path.join(ROOT_PATH, 'datapackage_pipelines', 'lib', processor.strip() + '.py')


for filename, testfunc in StdlibfixtureTests(os.path.join(os.path.dirname(__file__), 'fixtures')).get_tests():
    globals()['test_stdlib_%s' % filename] = testfunc




import os

from celery import Celery
from celery.schedules import crontab

from datapackage_pipelines.specs import pipelines, register_all_pipelines

import logging

REGULAR_TASK_NAME = 'datapackage_pipelines.celery_tasks.celery_tasks' + \
                        '.execute_pipeline_task'
SCHEDULED_TASK_NAME = 'datapackage_pipelines.celery_tasks.celery_tasks' + \
                        '.execute_scheduled_pipeline'

CELERY_SCHEDULE = {}

register_all_pipelines()

for spec in pipelines():
    if spec.schedule is not None:
        entry = {
            'task': SCHEDULED_TASK_NAME,
            'schedule': crontab(*spec.schedule),
            'args': (spec.pipeline_id,)
        }
        CELERY_SCHEDULE[spec.pipeline_id] = entry
        logging.info('SCHEDULING task %r: %r', spec.pipeline_id, spec.schedule)

celery_app = Celery('dpp')
celery_app.conf.update(CELERYBEAT_SCHEDULE=CELERY_SCHEDULE,
                       CELERY_TIMEZONE='UTC',
                       CELERY_REDIRECT_STDOUTS=False,
                       BROKER_URL=os.environ.get('CELERY_BROKER', 'amqp://'),
                       CELERY_RESULT_BACKEND=os.environ.get('CELERY_BACKEND',
                                                            'amqp://'),
                       CELERYD_LOG_LEVEL="INFO",
                       CELERY_TASK_SERIALIZER='json',
                       CELERY_RESULT_SERIALIZER='json',
                       CELERY_ACCEPT_CONTENT=['json'],
                       CELERY_ROUTES={
                           REGULAR_TASK_NAME: {'queue': 'datapackage-pipelines'},
                           SCHEDULED_TASK_NAME: {'queue': 'datapackage-pipelines'},
                       })



import datapackage
from datapackage.exceptions import DataPackageException

from ..errors import SpecError
from ...status import status


class DependencyMissingException(Exception):
    pass


def resolve_dependencies(spec, all_pipeline_ids):

    cache_hash = ''
    dependencies = spec.pipeline_details.get('dependencies', ())
    for dependency in dependencies:
        if 'pipeline' in dependency:
            pipeline_id = dependency['pipeline']
            if pipeline_id not in all_pipeline_ids:
                raise DependencyMissingException()

            dirty = all_pipeline_ids.get(pipeline_id).dirty
            if dirty is True:
                spec.errors.append(
                    SpecError('Dirty dependency',
                              'Cannot run until all dependencies are executed')
                )
            if not status.is_successful(pipeline_id):
                spec.errors.append(
                    SpecError('Dependency unsuccessful',
                              'Cannot run until all dependencies are '
                              'successfully executed')
                )
            pipeline_hash = all_pipeline_ids.get(pipeline_id).cache_hash
            assert pipeline_hash is not None
            cache_hash += pipeline_hash

            spec.dependencies.append(pipeline_id)

        elif 'datapackage' in dependency:
            dp_id = dependency['datapackage']
            try:
                dp = datapackage.DataPackage(dp_id)
                if 'hash' in dp.descriptor:
                    cache_hash += dp.descriptor['hash']
                else:
                    spec.errors.append(
                        SpecError('Missing dependency',
                                  "Couldn't get data from datapackage %s"
                                  % dp_id))
            except DataPackageException:
                spec.errors.append(
                    SpecError('Missing dependency',
                              "Couldn't open datapackage %s"
                              % dp_id))

        else:
            spec.errors.append(
                SpecError('Missing dependency',
                          'Unknown dependency provided (%r)' % dependency))

    return cache_hash



from datapackage_pipelines.wrapper import ingest, spew

parameters, datapackage, res_iter = ingest()


if datapackage is None:
    datapackage = {}

datapackage.setdefault('resources', [])

assert 'path' not in parameters, \
    "You can only add remote resources using this processor"
for param in ['url', 'name']:
    assert param in parameters, \
        "You must define {} in your parameters".format(param)

datapackage['resources'].append(parameters)

spew(datapackage, res_iter)



import os
import csv
import tempfile
import logging
import hashlib

from jsontableschema.exceptions import InvalidCastError
from jsontableschema.model import SchemaModel

from ...utilities.extended_json import json
from ...wrapper import ingest, spew


class DumperBase(object):

    def __init__(self):
        self.__params, self.__datapackage, self.__res_iter = ingest()
        self.stats = {}

    def __call__(self):
        self.initialize(self.__params)
        self.__datapackage = \
            self.prepare_datapackage(self.__datapackage, self.__params)
        spew(self.__datapackage,
             self.handle_resources(self.__datapackage,
                                   self.__res_iter,
                                   self.__params,
                                   self.stats),
             self.stats)
        self.handle_datapackage(self.__datapackage, self.__params, self.stats)
        self.finalize()

    def prepare_datapackage(self, datapackage, _):
        return datapackage

    @staticmethod
    def schema_validator(resource):
        schema = SchemaModel(resource.spec['schema'])
        for row in resource:
            for k, v in row.items():
                try:
                    schema.cast(k, v)
                except InvalidCastError:
                    logging.error('Bad value %r for field %s', v, k)
                    raise
            yield row

    @staticmethod
    def row_counter(datapackage, resource_spec, resource):
        resource_spec['count_of_rows'] = 0
        for row in resource:
            datapackage['count_of_rows'] += 1
            resource_spec['count_of_rows'] += 1
            if datapackage['count_of_rows'] % 1 == 10000:
                logging.info('Dumped %d rows', datapackage['count_of_rows'])
            yield row

    @staticmethod
    def hasher(datapackage, resource_spec, resource):
        resource_spec['hash'] = hashlib.md5()
        for row in resource:
            row_dump = json.dumps(row,
                                  sort_keys=True,
                                  ensure_ascii=True)\
                           .encode('utf8')
            resource_spec['hash'].update(row_dump)
            datapackage['hash'].update(row_dump)
            yield row
        resource_spec['hash'] = resource_spec['hash'].hexdigest()

    def handle_resources(self, datapackage,
                         resource_iterator,
                         parameters, stats):
        datapackage['count_of_rows'] = 0
        datapackage['hash'] = hashlib.md5()
        for resource in resource_iterator:
            resource_spec = resource.spec
            ret = self.handle_resource(DumperBase.schema_validator(resource),
                                       resource_spec,
                                       parameters,
                                       datapackage)
            ret = DumperBase.row_counter(datapackage, resource_spec, ret)
            ret = DumperBase.hasher(datapackage, resource_spec, ret)
            yield ret

        datapackage['hash'] = datapackage['hash'].hexdigest()
        stats['count_of_rows'] = datapackage['count_of_rows']
        stats['dataset_name'] = datapackage['name']

    def handle_datapackage(self, datapackage, parameters, stats):
        pass

    def handle_resource(self, resource, spec, parameters, datapackage):
        raise NotImplementedError()

    def initialize(self, params):
        pass

    def finalize(self):
        pass


PYTHON_DIALECT = {
    'number': {
        'decimalChar': '.',
        'groupChar': ''
    },
    'date': {
        'format': 'fmt:%Y-%m-%d'
    },
    'time': {
        'format': 'fmt:%H:%M:%S.%f'
    },
    'datetime': {
        'format': '%Y-%m-%d %H:%M:%S.%f'
    },
}

SERIALIZERS = {
    'array': json.dumps,
    'object': json.dumps,
}


class CSVDumper(DumperBase):

    def prepare_datapackage(self, datapackage, params):
        datapackage = \
            super(CSVDumper, self).prepare_datapackage(datapackage, params)

        # Make sure all resources are proper CSVs
        for resource in datapackage['resources']:
            resource['encoding'] = 'utf-8'
            basename, _ = os.path.splitext(resource['path'])
            resource['path'] = basename + '.csv'
            resource['format'] = 'csv'
            resource['dialect'] = dict(
                lineTerminator='\r\n',
                delimiter=',',
                doubleQuote=True,
                quoteChar='"',
                skipInitialSpace=False
            )
            for field in resource.get('schema', {}).get('fields', []):
                field.update(PYTHON_DIALECT.get(field['type'], {}))

        return datapackage

    def handle_datapackage(self, datapackage, parameters, stats):
        temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        json.dump(datapackage, temp_file, sort_keys=True, ensure_ascii=True)
        temp_file_name = temp_file.name
        temp_file.close()
        self.write_file_to_output(temp_file_name, 'datapackage.json')

    def write_file_to_output(self, filename, path):
        raise NotImplementedError()

    def rows_processor(self, resource, spec, _csv_file, _writer, _fields):
        for row in resource:
            transformed_row = CSVDumper.__transform_row(row, _fields)
            _writer.writerow(transformed_row)
            yield row
        filename = _csv_file.name
        _csv_file.close()
        self.write_file_to_output(filename, spec['path'])

    def handle_resource(self, resource, spec, _, datapackage):
        schema = spec['schema']

        temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
        fields = schema['fields']
        headers = list(map(lambda field: field['name'], fields))

        csv_writer = csv.DictWriter(temp_file, headers)
        csv_writer.writeheader()

        fields = dict((field['name'], field) for field in fields)

        return self.rows_processor(resource,
                                   spec,
                                   temp_file,
                                   csv_writer,
                                   fields)

    @staticmethod
    def __transform_value(value, field_type):
        if value is None:
            return ''
        serializer = SERIALIZERS.get(field_type, str)
        return serializer(value)

    @staticmethod
    def __transform_row(row, fields):
        return dict((k, CSVDumper.__transform_value(v, fields[k]['type']))
                    for k, v in row.items())
